/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <errno.h>
#include <systemd/sd-id128.h>

#include "common.h"
#include "event.h"
#include "log.h"
#include "module.h"

struct event_factory {
	struct faultd_module module;
	struct list_head types;
};

#define to_event_factory(MOD)						\
	container_of(MOD, struct event_factory, module)

static int init_event_factory(struct faultd_module *module,
							  struct faultd_config *config,
                              sd_event *event_loop)
{
	return 0;
}

static void cleanup_event_factory(struct faultd_module *module)
{
	struct event_factory *efactory = to_event_factory(module);

	assert(list_empty(&efactory->types));
}

static struct event_factory event_factory = {
	.module = {
		.name = "event_factory",
		.type = FAULTD_MODULE_TYPE_CORE,

		.init = init_event_factory,
		.cleanup = cleanup_event_factory,
		.node = LIST_HEAD_INIT(event_factory.module.node),
	},
	.types = LIST_HEAD_INIT(event_factory.types),
};

FAULTD_MODULE_REGISTER(&event_factory.module);

int faultd_event_type_register(struct faultd_event_type *new)
{
	if (!new->name || !new->allocate_event)
		return -EINVAL;

	list_add_tail(&new->node, &event_factory.types);

	return 0;
}

void faultd_event_type_unregister(struct faultd_event_type *type)
{
	if (list_empty(&type->node))
		return;

	list_del_init(&type->node);
}

int faultd_event_create(const char *type, void *data, struct faultd_event **ev)
{
	struct faultd_event_type *ev_type;
	bool from_bson;
	int ret = -EINVAL;

	if (!type)
		return ret;

	from_bson = (strcmp(type, BSON_EVENT_ID) == 0);
	if (from_bson) {
		struct faultd_object *d = data;
		/* get event type name from bson */
		ret = faultd_object_get_string(d, EV_TYPE_NAME, (char **)&type);
		if (ret < 0)
			return ret;
	}

	ret = -ENOENT;
	list_for_each_entry(ev_type, &event_factory.types, node)
		if (strcmp(ev_type->name, type) == 0)
			break;
	if (&ev_type->node == &event_factory.types)
		return ret;

	if (from_bson) {
		if (ev_type->deserialize_event == NULL) {
			ret = -ENOTSUP;
			log_error("Event type \"%s\" does not support deserialisation.",
					  ev_type->name);
		} else {
			ret = ev_type->deserialize_event(ev_type, data, ev);
			/* faultd_event_init_internal here? (old boot_id and oid) */
		}
	} else {
		ret = ev_type->allocate_event(ev_type, data, ev);
		/* faultd_event_init_internal here? (new boot_id and oid) */
	}
	return ret;
}

static void release_faultd_event(struct uref *uref)
{
	struct faultd_event *ev = container_of(uref, struct faultd_event, uref);

	if (ev->ops.release)
		ev->ops.release(ev);
}

int faultd_event_init_internal(struct faultd_event_type *ev_type,
							   struct faultd_event *ev)
{
	int ret;

	ret = clock_gettime(CLOCK_MONOTONIC, &ev->timestamp);
	if (ret)
		return ret;

	ret = sd_id128_get_boot(&ev->boot_id);
	if (ret < 0)
		return ret;

	ev->type = ev_type;
	ev->ops = ev_type->default_ops;

	uref_init(&ev->uref, release_faultd_event);
	INIT_LIST_HEAD(&ev->node);
	INIT_NQUEUE_NODE(&ev->nq_node);

	return 0;
}

void faultd_event_serialize_internal(struct faultd_event *ev, struct faultd_object *out)
{
	/* FIXME: check return values of faultd_object_append_XXX */
	faultd_object_append_uuid(out, EV_BOOT_ID, &ev->boot_id);
	/* FIXME: serialise event_type structure (?) */
	faultd_object_append_string(out, EV_TYPE_NAME, ev->type->name);
	faultd_object_append_timespec(out, EV_TIMESTAMP, &ev->timestamp);
}

int faultd_event_deserialize_internal(const struct faultd_object* in, struct faultd_event_type *ev_type, struct faultd_event *ev)
{
	struct faultd_object *obj;

	ev->type = ev_type;
	ev->ops = ev_type->default_ops;

	list_for_each_entry(obj, &in->val.children, node) {
		if (obj->type == TYPE_UUID && strcmp(EV_BOOT_ID, obj->key) == 0)
			memcpy(&ev->boot_id, &obj->val.uuid, sizeof(ev->boot_id));
		else if (obj->type == TYPE_TIMESPEC && strcmp(EV_TIMESTAMP, obj->key) == 0)
			memcpy(&ev->timestamp, &obj->val.ts, sizeof(ev->timestamp));
	}

	return 0;
}
