/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define _GNU_SOURCE 1

#include <assert.h>
#include <errno.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <systemd/sd-bus.h>
#include <json-c/json.h>
#include <unistd.h>

#include "log.h"
#include "service.h"
#include "common.h"
#include "systemd_dbus.h"

int systemd_service_init_by_pid(pid_t pid, struct systemd_service *s)
{
	_cleanup_free_ const char *n = NULL;
	int rc;

	assert(s);

	rc = systemd_get_unit_by_pid(pid, &n);
	if (rc < 0)
		return rc;

	return systemd_service_init(n, s);
}

static int get_config_field(struct json_object *root,
							const char *key, char **value)
{
	struct json_object *node;
	const char *node_value;
	json_bool rc;
	int ret = 0;

	/* if there is no such node let's skip it */
	rc = json_object_object_get_ex(root, key, &node);
	if (!rc)
		return 0;

	if (!json_object_is_type(node, json_type_string)) {
		log_error("Option %s has to be a string", key);
		return -EINVAL;
	}

	node_value = json_object_get_string(node);

	*value = strdup(node_value);
	if (*value == NULL) {
		log_error("Failed to duplicate string");
		ret = -ENOMEM;
	}

	return ret;
}

static int parse_service_config(const char *config_path,
								struct systemd_service *s)
{
	struct json_object *root;
	int ret;

	root = json_object_from_file(config_path);
	if (is_error(root)) {
		log_error("Could not create json object from file %s", config_path);
		return -1;
	}

	ret = get_config_field(root, "RecoveryUnit", &s->recovery_unit);
	if (ret < 0)
		goto out;

	ret = get_config_field(root, "ServiceType", &s->service_type);

out:
	json_object_put(root);
	return ret;
}

static int get_service_config_path(const char *dbus_path, char **config_path)
{
	_cleanup_(sd_bus_error_free) sd_bus_error error = SD_BUS_ERROR_NULL;
	_cleanup_(sd_bus_unrefp) sd_bus *bus = NULL;
	_cleanup_free_ char *name = NULL;
	int ret;

	ret = faultd_acquire_systemd_bus(&bus);
	if (ret < 0) {
		log_error_errno(ret, "Failed to acquire the default system bus connection: %m.");
		return ret;
	}

	ret = sd_bus_get_property_string(bus,
									 "org.freedesktop.systemd1",
									 dbus_path,
									 "org.freedesktop.systemd1.Unit",
									 "Id",
									 &error,
									 &name);
	if (ret < 0) {
		log_error("Failed to get \"Id\" property of %s: %s",
				  dbus_path, error.message);
		return ret;
	}

	log_debug("Service Id: %s", name);

	ret = asprintf(config_path, FAULTD_SERVICES_CONFIG_PATH "/%s.conf", name);
	if (ret < 0) {
		log_error("Could not allocate service configuration path: %m");
		return -ENOMEM;
	}

	return 0;
}

int systemd_service_init(const char *dbus_path, struct systemd_service *s)
{
	_cleanup_free_ char *config_path = NULL;
	int ret;

	assert(s);
	assert(dbus_path);

	memset(s, 0, sizeof(*s));

	s->dbus_path = strdup(dbus_path);
	if (!s->dbus_path)
		return -ENOMEM;

	ret = get_service_config_path(dbus_path, &config_path);
	if (ret)
		goto cleanup_service;

	ret = access(config_path, F_OK | R_OK);
	/* if there is no config file then we are done */
	if (ret && errno == ENOENT)
		return 0;

	if (ret) {
		log_error("Failed to access service configuration path: %m");
		ret = -errno;
		goto cleanup_service;
	}

	/* file exist so we need to parse it */
	ret = parse_service_config(config_path, s);
	if (ret) {
		log_error_errno(ret, "Failed to parse service configuration");
		goto cleanup_service;
	}

	return 0;

cleanup_service:
	systemd_service_cleanup(s);
	return ret;
}

void systemd_service_cleanup(struct systemd_service *s)
{
	if (s) {
		free(s->dbus_path);
		free(s->service_type);
		free(s->recovery_unit);
	}
}

int systemd_get_unit_by_pid(pid_t pid, const char **name)
{
	_cleanup_(sd_bus_error_free) sd_bus_error error = SD_BUS_ERROR_NULL;
	_cleanup_(sd_bus_message_unrefp) sd_bus_message *m = NULL;
	_cleanup_(sd_bus_unrefp) sd_bus *bus = NULL;
	const char *unit;
	int rc;

	assert(name);

	rc = faultd_acquire_systemd_bus(&bus);
	if (rc < 0) {
		log_error_errno(rc, "Failed to acquire the defult system bus connection: %m.");
		return rc;
	}

	rc = sd_bus_call_method(bus,
							SYSTEMD_SERVICE,
							SYSTEMD_OBJ,
							SYSTEMD_MANAGER_INTERFACE,
							"GetUnitByPID",                     /* method */
							&error,                             /* place to store error */
							&m,                                 /* place to store answer */
							"u",                                /* input signature */
							pid);                               /* input argument */
	if (rc < 0) {
		log_error("Failed to issue \"GetUnitByPID\" method call: %s.", error.message);
		return rc;
	}

	rc = sd_bus_message_read(m, "o", &unit);
	if (rc < 0) {
		log_error_errno(rc, "Failed to parse response message: %m.");
		return rc;
	}

	*name = strdup(unit);

	return 0;
}

void systemd_service_serialize(struct systemd_service *s, struct faultd_object* out)
{
	struct faultd_object *child;
	int ret;

	ret = faultd_object_new(&child);
	if (ret < 0)
		return;

	faultd_object_append_string(child, SYSD_SERVICE_DBUS_PATH, s->dbus_path);
	if (s->service_type != NULL)
		faultd_object_append_string(child, SYSD_SERVICE_SERVICE_TYPE, s->service_type);
	if (s->recovery_unit != NULL)
		faultd_object_append_string(child, SYSD_SERVICE_RECOVERY_UNIT, s->recovery_unit);

	faultd_object_append_object(out, SYSD_SERVICE, child);

	faultd_object_unref(child);
}

int systemd_service_deserialize(const struct faultd_object* in, struct systemd_service *s)
{
	struct faultd_object *obj;
	int ret = -EINVAL;
	bool dbus_path = false;

	memset(s, 0, sizeof(*s));

	list_for_each_entry(obj, &in->val.children, node) {
		if ((obj->type == TYPE_STRING) &&
			(strcmp(SYSD_SERVICE_DBUS_PATH, obj->key) == 0)) {
			s->dbus_path = strdup(obj->val.s);
			if (s->dbus_path == NULL) {
				ret = -errno;
				goto cleanup;
			}
			dbus_path = true;
		} else if ((obj->type == TYPE_STRING) &&
				   (strcmp(SYSD_SERVICE_SERVICE_TYPE, obj->key) == 0)) {
			s->service_type = strdup(obj->val.s);
			if (s->service_type == NULL) {
				ret = -errno;
				goto cleanup;
			}
		} else if ((obj->type == TYPE_STRING) &&
				   (strcmp(SYSD_SERVICE_RECOVERY_UNIT, obj->key) == 0)) {
			s->recovery_unit = strdup(obj->val.s);
			if (s->recovery_unit == NULL) {
				ret = -errno;
				goto cleanup;
			}
		}
	}

	if (dbus_path)
		ret = 0;

	return ret;

cleanup:
	systemd_service_cleanup(s);
	return ret;
}
