/*
 * This file is part of faultd.
 *
 * Copyright © 2017 Samsung Electronics
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <malloc.h>

#include "system_booted_event.h"

static int allocate_sb_event(struct faultd_event_type *type,
                             void *data, struct faultd_event **ev)
{
	struct system_booted_event *sb_ev;
	struct sb_event_data *sb_ev_data = data;
	int ret;

	sb_ev = malloc(sizeof(*sb_ev));
	if (!sb_ev)
		return -ENOMEM;
	memset(sb_ev, 0, sizeof(*sb_ev));

	ret = faultd_event_init_internal(type, &sb_ev->event);
	if (ret)
		goto free_sb_ev;

	sb_ev->prev_boot_id = sb_ev_data->prev_boot_id;
	*ev = &sb_ev->event;

	return 0;
free_sb_ev:
	free(sb_ev);
	return ret;
}

static void sb_event_release(struct faultd_event *ev)
{
	struct system_booted_event *sb_ev=
		to_system_booted_event(ev);

	faultd_event_cleanup_internal(&sb_ev->event);
	free(sb_ev);
}

static char *sb_event_to_string(struct faultd_event *ev)
{
	struct system_booted_event *sb_ev =
		to_system_booted_event(ev);
	char *str;
	int ret;

	ret = asprintf(&str, "System booted event:"
				   " Time: %lld.%.9ld"
				   " Previous boot ID: %s"
				   " Current boot ID: %s",
				   (long long)ev->timestamp.tv_sec,
				   ev->timestamp.tv_nsec,
				   SD_ID128_CONST_STR(sb_ev->prev_boot_id),
				   SD_ID128_CONST_STR(ev->boot_id)
				   );

	return ret > 0 ? str : NULL;
}

static void sb_event_serialize(struct faultd_event *ev, struct faultd_object *out)
{
	struct system_booted_event *sb_ev=
		to_system_booted_event(ev);

	faultd_event_serialize_internal(ev, out);
	faultd_object_append_uuid(out, "prev_boot_id", &sb_ev->prev_boot_id);
}
TEST_ENTRY_POINT(sb_event_serialize);

static struct faultd_event_type system_booted_event_type = {
	.name = SYSTEM_BOOTED_EVENT_ID,
	.default_ops = {
		.release = sb_event_release,
		.serialize = sb_event_serialize,
		.to_string = sb_event_to_string,
	},
	.allocate_event = allocate_sb_event,
	.node = LIST_HEAD_INIT(system_booted_event_type.node),
};

FAULTD_EVENT_TYPE_REGISTER(system_booted_event_type, system_booted_et)
