#include "helpers.h"

void test_serialization(struct serialized_element *data, int size, struct faultd_object *parent)
{
	for (int i = 0; i < size; ++i) {
		switch (data[i].type) {
			case TYPE_OID:
				{
					GET_FAULTD_OBJECT_VAL(faultd_oid_t, val, parent, data[i].key, data[i].type);
					assert_oid_equal(&val, data[i].desired);
					break;
				}
			case TYPE_STRING:
				{
					GET_FAULTD_OBJECT_VAL(char*, val, parent, data[i].key, data[i].type);
					assert_string_equal(val, data[i].desired);
					break;
				}
			case TYPE_INT:
				{
					GET_FAULTD_OBJECT_VAL(int, val, parent, data[i].key, data[i].type);
					assert_int_equal(val, *(int*)(data[i].desired));
					break;
				}
			case TYPE_TIMESPEC:
				{
					GET_FAULTD_OBJECT_VAL(struct timespec, val, parent, data[i].key, data[i].type);
					assert_timespec_equal(&val, data[i].desired);
					break;
				}
			case TYPE_TIME_T:
				{
					GET_FAULTD_OBJECT_VAL(time_t, val, parent, data[i].key, data[i].type);
					assert_time_t_equal(val, *(time_t*)data[i].desired);
					break;
				}
			case TYPE_UUID:
				{
					GET_FAULTD_OBJECT_VAL(sd_id128_t, val, parent, data[i].key, data[i].type);
					assert_uuid_equal(&val, data[i].desired);
					break;
				}
			case TYPE_OBJECT:
				{
					struct faultd_object *child;
					int ret;

					ret = faultd_object_get_object(parent, data[i].key, &child);
					assert_int_equal(ret, 0);
					test_serialization(data[i].child, data[i].children_num, child);
					break;
				}
			default:
				fprintf(stderr, "Type '%d' is not supported (key: '%s')\n",
						data[i].type, data[i].key);
				fail();
				break;
		}
	}

	assert_int_equal(faultd_object_count_children(parent), size);
}

void assert_faultd_object_equal(struct faultd_object *actual,
								struct faultd_object *desired)
{
	struct faultd_object *child_actual;
	struct faultd_object *child_desired;

	child_actual = list_first_entry(&actual->val.children,
									struct faultd_object,
									node);

	child_desired = list_first_entry(&desired->val.children,
									 struct faultd_object,
									 node);

	while (&child_actual->node != &actual->val.children)
	{
		if (child_actual->type == TYPE_OBJECT) {
			assert_faultd_object_equal(child_actual, child_desired);
		} else {
			assert_int_equal(child_actual->type, child_desired->type);
			assert_data_equal(child_actual->type,
							  &child_actual->val, &child_desired->val);
		}

		child_actual = list_next_entry(child_actual, node);
		child_desired = list_next_entry(child_desired, node);
	}
}

void assert_data_equal(int type, void *actual, void *desired)
{
	switch (type) {
	case TYPE_OID:
		assert_oid_equal(actual, desired);
		break;
	case TYPE_STRING:
		assert_string_equal(*(char **)actual, *(char **)desired);
		break;
	case TYPE_INT:
		assert_int_equal(*(int*)(actual),
						 *(int*)(desired));
		break;
	case TYPE_TIMESPEC:
		assert_timespec_equal(actual, desired);
		break;
	case TYPE_TIME_T:
		assert_time_t_equal(*(time_t*)actual,
							*(time_t*)desired);
		break;
	case TYPE_UUID:
		assert_uuid_equal(actual, desired);
		break;
	default:
		fprintf(stderr, "Type '%d' is not supported\n", type);
		fail();
		break;
	}
}

int faultd_object_count_children(struct faultd_object *obj)
{
	struct faultd_object *child;
	int i = 0;

	list_for_each_entry(child, &obj->val.children, node)
		i++;

	return i;
}

static int long_cmp(const long actual, const long expected)
{
	return actual == expected;
}

static int long_equal_display_error(const LargestIntegralType actual,
									const LargestIntegralType expected)
{
	if (long_cmp((const long)actual, (const long)expected))
		return 1;

	fprintf(stderr, "%ld != %ld\n",
			(const long)actual, (const long)expected);
	return 0;
}

void assert_long_equal(const long actual, const long expected)
{
	if (long_equal_display_error(
			cast_to_largest_integral_type(actual),
			cast_to_largest_integral_type(expected)) == 0)
		fail();
}

void assert_time_t_equal(const time_t actual, const time_t expected)
{
	assert_long_equal((const long)actual, (const long)expected);
}

void assert_timespec_equal(const struct timespec *actual, const struct timespec *expected)
{
	assert_long_equal(actual->tv_sec, expected->tv_sec);
	assert_long_equal(actual->tv_nsec, expected->tv_nsec);
}

void assert_oid_equal(const faultd_oid_t *actual, const faultd_oid_t *expected)
{
	assert_memory_equal(actual, expected, sizeof(faultd_oid_t));
}

static int uuid_cmp(const sd_id128_t *actual, const sd_id128_t *expected)
{
	return sd_id128_equal(*actual, *expected);
}

static int uuid_equal_display_error(const LargestIntegralType actual,
									const LargestIntegralType expected)
{
	const sd_id128_t *act = (const sd_id128_t *)cast_to_pointer_integral_type(actual);
	const sd_id128_t *exp = (const sd_id128_t *)cast_to_pointer_integral_type(actual);

	if (uuid_cmp(act, exp))
		return 1;


	fprintf(stderr, SD_ID128_FORMAT_STR " != " SD_ID128_FORMAT_STR "\n",
			SD_ID128_FORMAT_VAL(*act),
			SD_ID128_FORMAT_VAL(*exp));
	return 0;
}

void assert_uuid_equal(const sd_id128_t *actual, const sd_id128_t *expected)
{
	if (uuid_equal_display_error(
			cast_to_largest_integral_type(actual),
			cast_to_largest_integral_type(expected)) == 0)
		fail();
}

char generate_char()
{
	/* generate random char in range 0x20-0x7E (' ' to '~') excluding '.' */
	char c;

	while (1) {
		c = ' ' + (rand() % 94);
		if (c != '.')
			break;
	}

	return c;
}

char *generate_string(int len)
{
	char *str;

	str = calloc(len + 1, sizeof(char));
	assert_non_null(str);

	for (int i = 0; i < len; ++i) {
		str[i] = generate_char();

	}
	str[len] = '\0';

	return str;
}

int generate_int()
{
	return rand();
}

time_t generate_time_t()
{
	return generate_int();
}

bson_oid_t generate_bson_oid_t()
{
	bson_oid_t oid;

	for (int i = 0; i < ARRAY_SIZE(oid.ints); ++i)
		oid.ints[i] = generate_int();
	return oid;
}

sd_id128_t generate_uuid()
{
	sd_id128_t uuid;

	assert_int_equal(sd_id128_randomize(&uuid), 0);
	return uuid;
}

faultd_oid_t generate_oid()
{
	faultd_oid_t oid;

	memset(&oid, 0, sizeof(oid));
	oid.uuid = generate_uuid();
	return oid;
}

struct timespec generate_timespec()
{
	struct timespec t;

	t.tv_sec = generate_int();
	t.tv_nsec = generate_int();
	return t;
}

void test_event_serialization(void **state)
{
	struct serialize_test_data *td = *state;
	struct faultd_object *parent;
	int ret;

	ret = faultd_object_new(&parent);
	assert_int_equal(ret, 0);

	td->serialize_func(td->ev, parent);

	test_serialization(td->test_data, td->size, parent);

	faultd_object_unref(parent);
}

struct serialized_element *serialized_elems_dup(
	struct serialized_element *elems, size_t size)
{
	struct serialized_element *new_elems;
	int i;

	new_elems = calloc(size, sizeof(*elems));
	assert_non_null(new_elems);

	memcpy(new_elems, elems, sizeof(*elems)*size);

	for (i = 0; i < size; ++i)
		if (new_elems[i].type == TYPE_OBJECT)
			new_elems[i].child = serialized_elems_dup(new_elems[i].child,
													  new_elems[i].children_num);

	return new_elems;
}

void serialized_elems_free(struct serialized_element *elems, size_t size)
{
	int i;

	for (i = 0; i < size; ++i)
		if (elems[i].type == TYPE_OBJECT)
			serialized_elems_free(elems[i].child, elems[i].children_num);
}

int setup_event_serialization(void **state,
							  struct faultd_event *(*alloc_event)(),
							  void (*generate_attrs)(
								  struct serialize_test_data *td),
							  serialize_func_t serialize_func)
{
	struct serialize_test_data *td;

	srand(time(NULL));

	td = calloc(1, sizeof(*td));
	assert_non_null(td);

	td->ev = alloc_event();

	generate_attrs(td);

	td->serialize_func =  serialize_func;

	*state = td;
	return 0;
}
